require 'date'
require 'bigdecimal'

require_relative 'reconciliation'
require_relative 'post'

module Rledger
  # Transaction base contains the basic elements of any transaction
  class TransactionBase
    attr_accessor :date, :id, :reconciliation, :payee

    # initialize variables with the correct types
    def initialize
      @date = Date.new
      @id = ""
      @reconciliation = Reconciliation.new
      @payee = ""
    end

    def to_s
      "#{@date.strftime("%d/%m/%Y")}#{rec_to_s}#{id_to_s}#{payee_to_s}\n"
    end
    
    def to_qif(account)
      "D#{@date}\n" +
        (@payee != "" ? "P#{@payee}\n" : "") +
        (@id != "" ? "N#{@id}\n" : "")
    end

    def parse(s)
      # regular expressions of various components of a transaction base
      date = "([0-9][0-9]/[0-9][0-9]/[0-9][0-9][0-9][0-9])"
      ob = "[\t ]*"
      rec = "([!?*]|)"
      id = "(\\([^)]+\\)|)"
      payee = "(.*)"

      match = Regexp.new(date + ob + rec + ob + id + ob + payee).match(s)
      if match
        @date = s_to_date(match[1])  # TODO: replace with Chronic o Date.parse
        @reconciliation.parse(match[2])
        @id = match[3] == "" ? "" : match[3][1..-2] # so that id is always a string
        @payee = match[4]
        true
      else
        false
      end
    end

    private
    
    def id_to_s
      @id == "" ? "" : "\t(#{@id})"
    end

    def rec_to_s
      @reconciliation.unknown? ? "" : "\t#{@reconciliation.to_s}"
    end

    def payee_to_s
      @payee == "" ? "" : "\t#{@payee}"
    end

    def s_to_date(s)
      day, month, year = s.split("/");
      Date.civil(year.to_i, month.to_i, day.to_i)
    end

  end

  # A transaction models a multiple entry transactions
  # It should contain a minimum of two entries 
  class Transaction < TransactionBase
    attr_accessor :posts

    def initialize
      super
      @posts = []
    end

    def self.read(filename)
      transactions = []
      file_string = IO.read(filename)
      array = file_string.split(/\n([\t ]*\n)+/)  # ignore empty lines (possibly with blanks)
      array.each do |record|
        if record =~ /^[0-9]/
          t = Transaction.new
          t.parse(record)
          transactions << t
        end
      end
      transactions
    end

    def <=> other
      date <=> other.date
    end

    def parse(s)
      lines = s.split("\n")
      super(lines[0]) # call super.parse!
      lines[1..-1].each do |line|
        p = Post.new
        p.parse(line)
        @posts << p
      end

      # fix the derived amount in the derived entry (if there is one)
      # TODO Raise a lot of errors:
      # - the other entries are not in a single currency 
      # - there is more than a derived entry
      p = @posts.select { |x| x.derived? }
      if p[0] != nil then
        p[0].amount = total_no_derived.multiply!(BigDecimal.new(-1))
      end
    end

    def total_no_derived
      a = Amount.new
      @posts.each do |e|
        a.add!(e.amount) if not e.derived?
      end
      a
    end

    def posts_with voice
      @posts.select { |e| e.voice == voice }
    end

    def posts_without voice
      @posts.select { |e| e.voice != voice }
    end

    def contains? voice
      posts_with(voice).size > 0
    end

    def to_s
      accumulator = ""
      @posts.each { |post| accumulator << post.to_s }
      super.to_s + accumulator + "\n"
    end
    
    def to_qif(account)
      accumulator = super.to_s   
      @posts.each { |post|
        accumulator <<  
        "L[#{post.item}]\n" +
        "$#{post.amount}\n" +
        "M#{post.comment}\n"
      }
      accumulator + "^"
    end
    
  end
end
